package quarkconfig

import corev1 "k8s.io/api/core/v1"

const (
	nobodyUserId  = int64(65534)
	nonrootUserId = int64(65532)
)

// getPodSecurityContext returns PodSecurityContext with provided as user
func getPodSecurityContext(runAsUser int64, runAsGroup int64) *corev1.PodSecurityContext {
	fsGroup := runAsGroup
	runAsNonRoot := true
	return &corev1.PodSecurityContext{
		RunAsUser:    &runAsUser,
		RunAsGroup:   &runAsGroup,
		RunAsNonRoot: &runAsNonRoot,
		FSGroup:      &fsGroup,
	}
}

func getComponentUserIdMap() map[string]int64 {
	return map[string]int64{
		CSCDeploymentName:               nonrootUserId, // cloud-snapshot-controller
		NVCDeploymentName:               nonrootUserId, // netapp-volume-controller
		NodevolControllerName:           nonrootUserId, // nodevol-controller
		SvcMeshControllerDeploymentName: nonrootUserId, // asmc - netapp-svcmesh-controller
		NodeMonitorDaemonSetName:        nonrootUserId, // node-monitor
		NHCStatefulSetName:              nobodyUserId,  // netapp-ha-controller

		// TODO: QUARK-4125: nodevol-pod or nodevol-worker
		// nodevol-worker
		// TODO: Below components needs to be migrated to distroless nonroot base image
		// TODO: QUARK-4124
		//NSCDaemonSetName:                    nobodyUserId,  // netapp-snat-controller

		EtcdOperatorDeploymentName:          nonrootUserId, // etcd-operator
		DefaultPrepullDaemonSet:             nobodyUserId,  // image-prepull
		QuarkEtcdClusterName:                nobodyUserId,  // quark-etcd-cluster
		KubeStateMetricsDeploymentName:      nobodyUserId,  // quark-kube-state-metrics
		QuarkOtelDeploymentName:             nobodyUserId,  // quark-otel-collector
		QuarkPrometheusServerDeploymentName: nobodyUserId,  // quark-prometheus-server
		QuarkContainerPromReload:            nonrootUserId, // prometheus-server-configmap-reload

		// quark pod containers
		QuarkContainerPodrick: nobodyUserId,
		QuarkContainerLogging: nobodyUserId,

		// Not yet supported
		// TODO: Enable non-root support for below containers and deployments
		// HASpareDeploymentName:        nobodyUserId,
		// QuarkContainerDMAP:           nobodyUserId,
		// QuarkContainerSupportability: nobodyUserId,
		// QuarkContainerSECD:           nobodyUserId,
		// QuarkContainerCCPD:           nobodyUserId,

		// INFO: quark-operator security context has to be updated in the helm chart
		// INFO: etcd-cluster -- this should go to etcd-operator repo
	}
}

func GetPodSecurityContext(podName string) *corev1.PodSecurityContext {
	uidMap := getComponentUserIdMap()
	if uid, ok := uidMap[podName]; ok {
		return getPodSecurityContext(uid, uid)
	}
	return nil
}

// GetAtomQuarkPodSecurityContext - to get the atom-quark pod security context
func GetAtomQuarkPodSecurityContext() *corev1.PodSecurityContext {
	return getPodSecurityContext(nobodyUserId, nobodyUserId)
}

// getContainerSecurityContext returns container level Security Contexts that containers should be configured with
func getContainerSecurityContext(runAsUser, runAsGroup int64) *corev1.SecurityContext {
	privileged := false
	readOnlyRootFilesystem := true
	allowPrivilegeEscalation := false
	dropAllCapabilities := []corev1.Capability{"ALL"}
	runAsNonRoot := true
	return &corev1.SecurityContext{
		Privileged:               &privileged,
		ReadOnlyRootFilesystem:   &readOnlyRootFilesystem,
		AllowPrivilegeEscalation: &allowPrivilegeEscalation,
		RunAsUser:                &runAsUser,
		RunAsGroup:               &runAsGroup,
		RunAsNonRoot:             &runAsNonRoot,
		Capabilities: &corev1.Capabilities{
			Drop: dropAllCapabilities,
		},
	}
}

// GetContainerSecurityContext returns container level Security Contexts that containers should be configured with
func GetContainerSecurityContext(compoentName string) *corev1.SecurityContext {
	uidMap := getComponentUserIdMap()
	if userId, ok := uidMap[compoentName]; ok {
		return getContainerSecurityContext(userId, userId)
	}
	return nil
}
