/*
 *  Copyright (c) 2023 NetApp
 *  All rights reserved
 */
package util

import (
	"context"
	"fmt"
	"strings"
	"sync"

	appsv1 "k8s.io/api/apps/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	types "k8s.io/apimachinery/pkg/types"
	"k8s.io/client-go/kubernetes"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// GetSecretResourceVersion returns secret's resource version
func GetSecretResourceVersion(k8sclient kubernetes.Interface, namespace, secretName string) (string, error) {
	secretResource, err := k8sclient.CoreV1().Secrets(namespace).Get(context.TODO(), secretName, metav1.GetOptions{})
	if err != nil {
		return "", fmt.Errorf("unable to get secret %s in namespace %s. err: %s", secretName, namespace, err.Error())
	}
	return secretResource.ResourceVersion, nil
}

// GetSecretResourceAnnotationKey returns annotation key for the secret
func GetSecretResourceAnnotationKey(secretName string) string {
	return fmt.Sprintf("secrets/%s", secretName)
}

func UpdateStatefulSetWithCertSecretAnnotation(k8sclient kubernetes.Interface, sts *appsv1.StatefulSet, secretNames []string) error {
	namespace := sts.Namespace
	for _, secretName := range secretNames {
		resourceVersion, err := GetSecretResourceVersion(k8sclient, namespace, secretName)
		if err != nil {
			return err
		}
		if sts.Spec.Template.Annotations == nil {
			sts.Spec.Template.Annotations = map[string]string{}
		}
		sts.Spec.Template.Annotations[GetSecretResourceAnnotationKey(secretName)] = resourceVersion
	}
	return nil
}

// UpdateDeploymentWithCertSecretAnnotation updates deployment with the annotation for given secrets
func UpdateDeploymentWithCertSecretAnnotation(k8sclient kubernetes.Interface, depl *appsv1.Deployment, secretNames []string) error {
	namespace := depl.Namespace
	for _, secretName := range secretNames {
		resourceVersion, err := GetSecretResourceVersion(k8sclient, namespace, secretName)
		if err != nil {
			return err
		}
		if depl.Spec.Template.Annotations == nil {
			depl.Spec.Template.Annotations = map[string]string{}
		}
		depl.Spec.Template.Annotations[GetSecretResourceAnnotationKey(secretName)] = resourceVersion
	}
	return nil
}

func UpdateDaemonSetWithCertSecretAnnotation(k8sclient kubernetes.Interface, dms *appsv1.DaemonSet, secretNames []string) error {
	namespace := dms.Namespace
	for _, secretName := range secretNames {
		resourceVersion, err := GetSecretResourceVersion(k8sclient, namespace, secretName)
		if err != nil {
			return err
		}
		if dms.Spec.Template.Annotations == nil {
			dms.Spec.Template.Annotations = map[string]string{}
		}
		dms.Spec.Template.Annotations[GetSecretResourceAnnotationKey(secretName)] = resourceVersion
	}
	return nil
}

// names of resources to be reloaded when there is secret change
const (
	// applications/services from storage/nvc namespace
	CertReloadAppsNHC              string = quarkconfig.NHCStatefulSetName
	CertReloadAppsNVC              string = quarkconfig.NVCDeploymentName
	CertReloadAppsQuarkDeployments string = "quark-deployments"
	// applications/services from quark-system namespace
	CertReloadAppsNodeMonitor  string = quarkconfig.NodeMonitorDaemonSetName
	CertReloadAppsEtcdOperator string = quarkconfig.EtcdOperatorDeploymentName
	CertReloadAppsEtcdCluster  string = quarkconfig.QuarkEtcdClusterName
)

// dataMapAppsForCertSecret stores list of applications using the secret in a namespace
// map[namespace/secret][]appNames
var dataMapAppsForCertSecret map[string][]string

// dataMapCertSecretsForApp stores list of secretnames used by the app
// map[appName][]namespace/secret -
var dataMapCertSecretsForApp map[string][]string

var certSecretMapLock = &sync.Mutex{}

func getNamespacedName(name, namespace string) string {
	return types.NamespacedName{Name: name, Namespace: namespace}.String()
}

func getNamespacedSecretNames(namespace string, secrets []string) []string {
	result := []string{}
	for _, secretName := range secrets {
		result = append(result, getNamespacedName(secretName, namespace))
	}
	return result
}

// InitializeCertSecretDataMap initializes hash map that stores secrets to applications.
// argument: netappVolNamespace is the namespace used for quark pod / nvc deployments
func InitializeCertSecretDataMap(netappVolNamespace string) {
	certSecretMapLock.Lock()
	defer certSecretMapLock.Unlock()
	if dataMapCertSecretsForApp != nil && len(dataMapCertSecretsForApp) > 0 {
		return
	}
	// Build maps with key:namespace/appNames  value: list of string formatted namespace/secret-name
	dataMapCertSecretsForApp = map[string][]string{
		getNamespacedName(CertReloadAppsNHC, netappVolNamespace): getNamespacedSecretNames(netappVolNamespace, []string{
			quarkconfig.SecretNameQuarkTrustedCA,
			quarkconfig.SecretNameHaHbServerTls,
		}),
		getNamespacedName(CertReloadAppsNVC, netappVolNamespace): getNamespacedSecretNames(netappVolNamespace, []string{
			quarkconfig.SecretNameQuarkTrustedCA,
			quarkconfig.SecretNameEtcdClientTls,
		}),
		getNamespacedName(CertReloadAppsQuarkDeployments, netappVolNamespace): getNamespacedSecretNames(netappVolNamespace, []string{
			quarkconfig.SecretNameQuarkTrustedCA,
			quarkconfig.SecretNameEtcdClientTls,
		}),
		getNamespacedName(CertReloadAppsNodeMonitor, quarkconfig.QuarkSystemNamespace): getNamespacedSecretNames(quarkconfig.QuarkSystemNamespace, []string{
			quarkconfig.SecretNameQuarkTrustedCA,
			quarkconfig.SecretNameHaHbClientTls,
		}),
		getNamespacedName(CertReloadAppsEtcdOperator, quarkconfig.QuarkSystemNamespace): getNamespacedSecretNames(quarkconfig.QuarkSystemNamespace, []string{
			quarkconfig.SecretNameQuarkTrustedCA,
			quarkconfig.SecretNameEtcdOperatorTls,
		}),
		getNamespacedName(CertReloadAppsEtcdCluster, quarkconfig.QuarkSystemNamespace): getNamespacedSecretNames(quarkconfig.QuarkSystemNamespace, []string{
			quarkconfig.SecretNameQuarkTrustedCA,
			quarkconfig.SecretNameEtcdPeerTls,
			quarkconfig.SecretNameEtcdServerTls,
			quarkconfig.SecretNameEtcdOperatorTls,
		}),
	}
	// Build map with key: namespace/secret-name value: list of appNames
	dataMapAppsForCertSecret = map[string][]string{}
	// Build maps with key:appNames  value: list of string formatted namespace/secret-name
	for namespacedAppName, namespacedSecrets := range dataMapCertSecretsForApp {
		for _, namespacedSecret := range namespacedSecrets {
			if _, ok := dataMapAppsForCertSecret[namespacedSecret]; !ok {
				dataMapAppsForCertSecret[namespacedSecret] = []string{}
			}
			dataMapAppsForCertSecret[namespacedSecret] = append(dataMapAppsForCertSecret[namespacedSecret], namespacedAppName)
		}
	}
}

// getUniqueStringSlice returns slice containing unique strings
func getUniqueStringSlice(inputSlice []string) []string {
	keys := make(map[string]bool)
	result := []string{}
	for _, entry := range inputSlice {
		if _, value := keys[entry]; !value {
			keys[entry] = true
			result = append(result, entry)
		}
	}
	return result
}

// SplitNamespacedName returns strings in the order of namespace, name and result.
// result is true if split is successful, otherwise false
func SplitNamespacedName(namespacedName string) (string, string, bool) {
	tokens := strings.Split(namespacedName, string(types.Separator))
	if len(tokens) != 2 {
		return "", "", false
	}
	return tokens[0], tokens[1], true
}

// GetCertAppDataListForSecret returns list of application names for the given secret filtered by namespace.
// if namespace is empty, returns apps that uses secret from all namespaces
func GetCertAppDataListForSecret(secret, namespace string) []string {
	if namespace != "" {
		nsSecret := getNamespacedName(secret, namespace)
		retval := dataMapAppsForCertSecret[nsSecret]
		return retval
	}
	result := []string{}
	for namespacedSecretName, namespacedAppNames := range dataMapAppsForCertSecret {
		if _, secreName, ok := SplitNamespacedName(namespacedSecretName); ok && secreName == secret {
			result = append(result, namespacedAppNames...)
		}
	}
	return getUniqueStringSlice(result)
}

// GetSecretCertAppDataForService returns list of namespaced secret names for the given application filtered
// by namespace. if namespace is empty, returns secrets from all namespaces which are consumed by this app.
func GetSecretCertAppDataForService(appName, namespace string) []string {
	if namespace != "" {
		nsName := getNamespacedName(appName, namespace)
		return dataMapCertSecretsForApp[nsName]
	}
	result := []string{}
	for namespacedAppName, namespacedSecretNames := range dataMapCertSecretsForApp {
		if _, name, ok := SplitNamespacedName(namespacedAppName); ok && name == appName {
			result = append(result, namespacedSecretNames...)
		}
	}
	return getUniqueStringSlice(result)
}

// IsStringInSlice returns index position of the string in the slice.
// negative return value denote string not found
func IsStringInSlice(str string, list []string) int {
	for i, item := range list {
		if item == str {
			return i
		}
	}
	return -1
}

// IsCertSecretInNetAppVolumeNamespace returns true of the given secret is present in the
// storage namespace (nvc namespace)
func IsCertSecretInNetAppVolumeNamespace(lookupSecret string) bool {
	return IsStringInSlice(lookupSecret, quarkconfig.NVCNamespaceSecrets) >= 0
}

// IsCertSecretInQuarkSystemNamespace returns true of the given secret is present in the
// quark-system namespace
func IsCertSecretInQuarkSystemNamespace(lookupSecret string) bool {
	return IsStringInSlice(lookupSecret, quarkconfig.QuarkSystemNamespaceSecrets) >= 0
}

// UpdateNHCStatefulSetWithCertSecretAnnotation Set NHC Statefulset with the secrets annotation used for certificates
func UpdateNHCStatefulSetWithCertSecretAnnotation(k8sclient kubernetes.Interface, sts *appsv1.StatefulSet) error {
	secretNames := []string{quarkconfig.SecretNameHaHbServerTls, quarkconfig.SecretNameQuarkTrustedCA}
	return UpdateStatefulSetWithCertSecretAnnotation(k8sclient, sts, secretNames)
}

// UpdateNodeMonitorDaemonSetWithCertSecretAnnotation Node Monitor DaemonSet
func UpdateNodeMonitorDaemonSetWithCertSecretAnnotation(k8sclient kubernetes.Interface, dms *appsv1.DaemonSet) error {
	secretNames := []string{quarkconfig.SecretNameHaHbClientTls, quarkconfig.SecretNameQuarkTrustedCA}
	return UpdateDaemonSetWithCertSecretAnnotation(k8sclient, dms, secretNames)
}

// UpdateNVCDeploymentWithCertSecretAnnotation NVC Deployment
func UpdateNVCDeploymentWithCertSecretAnnotation(k8sclient kubernetes.Interface, depl *appsv1.Deployment) error {
	secretNames := []string{quarkconfig.SecretNameEtcdClientTls, quarkconfig.SecretNameQuarkTrustedCA}
	return UpdateDeploymentWithCertSecretAnnotation(k8sclient, depl, secretNames)
}

// UpdateQuarkDeploymentWithCertSecretAnnotation Quark Deployment
func UpdateQuarkDeploymentWithCertSecretAnnotation(k8sclient kubernetes.Interface, depl *appsv1.Deployment) error {
	secretNames := []string{quarkconfig.SecretNameEtcdClientTls, quarkconfig.SecretNameQuarkTrustedCA}
	return UpdateDeploymentWithCertSecretAnnotation(k8sclient, depl, secretNames)
}
