package helm

import (
	"context"
	"fmt"
	"os"
	"time"

	log "github.com/sirupsen/logrus"
	k8serrors "k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/meta"
	"k8s.io/client-go/discovery"
	"k8s.io/client-go/discovery/cached/memory"
	"k8s.io/client-go/restmapper"
	"k8s.io/client-go/tools/clientcmd"

	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
	apierrors "k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
)

// MigrateQuarkResourcesToHelm takes a kubeconfig, quark name and quark namespace and adds annotations to existing quark resources
// to allow quark to be upgraded with the helm workflow
func (h *HelmReconciler) migrateQuarkResourcesToHelm() error {
	log.Debugln("Migrating resources to helm")
	resources := []struct {
		Resource    string
		Name        string
		ReleaseName string
		Namespace   string
	}{
		{ResourceQuark, GetQuarkCRName(h.Values, h.Namespace), "quark", h.Namespace},
		{ResourceDeployment, "quark-operator", "quark-operator", h.Namespace},
		{ResourceClusterRoleBinding, "quark-operator-asm", "quark-operator", h.Namespace},
		{ResourceClusterRoleBinding, "quark-operator", "quark-operator", h.Namespace},
		{ResourceClusterRole, "quark-operator", "quark-operator", h.Namespace},
		{ResourceClusterRole, "netapp-asmc-crole", "quark-operator", h.Namespace},
		{ResourceServiceAccount, "quark-operator", "quark-operator", h.Namespace},
		{ResourceCrd, "quarks.quark.netapp.io", "quark", h.Namespace},
	}

	for _, r := range resources {
		err := AnnotateAndLabelK8sResource(h.Kubeconfig, r.Resource, r.Name, r.ReleaseName, r.Namespace)
		if err != nil {
			if k8serrors.IsNotFound(err) {
				log.Debugf("Could not find resource %s, skipping", r.Name)
			} else {
				return err
			}
		}
	}
	return nil
}

// AnnotateAndLabelK8sResource annotates an existing k8s resource with a specific helm release name, so helm is aware of it for upgrade/install
func AnnotateAndLabelK8sResource(config *rest.Config, resource, name, releaseName, namespace string) error {
	log.WithFields(log.Fields{
		"resource":  resource,
		"name":      name,
		"namespace": namespace,
	}).Debugln("Running AnnotateAndLabelK8sResource")

	quarkClient, err := quarkclientset.NewForConfig(config)
	if err != nil {
		return err
	}

	apiextensionsclientset, err := apiextensionsclient.NewForConfig(config)
	if err != nil {
		return err
	}

	clientset, err := kubernetes.NewForConfig(config)
	if err != nil {
		return err
	}

	return annotateAndLabelK8sResource(quarkClient, apiextensionsclientset, clientset, resource, name, releaseName, namespace)
}

// annotateAndLabelK8sResource annotates an existing k8s resource with a specific helm release name, so helm is aware of it for upgrade/install
func annotateAndLabelK8sResource(quarkClient quarkclientset.Interface, apiExClient apiextensionsclient.Interface, kubeClientset kubernetes.Interface, resource, name, releaseName, namespace string) error {

	ctx := context.Background()
	var updateFunc func(context.Context, metav1.UpdateOptions) (metav1.Object, error)

	switch resource {
	case ResourceDeployment:

		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			deployment, err := kubeClientset.AppsV1().Deployments(namespace).Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			applyAnnotationsAndLabels(deployment, releaseName, namespace)
			return kubeClientset.AppsV1().Deployments(namespace).Update(ctx, deployment, opts)
		}
	case ResourceClusterRoleBinding:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			clusterRoleBinding, err := kubeClientset.RbacV1().ClusterRoleBindings().Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			applyAnnotationsAndLabels(clusterRoleBinding, releaseName, namespace)
			return kubeClientset.RbacV1().ClusterRoleBindings().Update(ctx, clusterRoleBinding, opts)
		}
	case ResourceClusterRole:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			clusterRole, err := kubeClientset.RbacV1().ClusterRoles().Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			applyAnnotationsAndLabels(clusterRole, releaseName, namespace)
			return kubeClientset.RbacV1().ClusterRoles().Update(ctx, clusterRole, opts)
		}
	case ResourceServiceAccount:

		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			serviceAccount, err := kubeClientset.CoreV1().ServiceAccounts(namespace).Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			applyAnnotationsAndLabels(serviceAccount, releaseName, namespace)
			return kubeClientset.CoreV1().ServiceAccounts(namespace).Update(ctx, serviceAccount, opts)
		}
	case ResourceCrd:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			customResourceDefinition, err := apiExClient.ApiextensionsV1().CustomResourceDefinitions().Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			applyAnnotationsAndLabels(customResourceDefinition, releaseName, namespace)
			return apiExClient.ApiextensionsV1().CustomResourceDefinitions().Update(ctx, customResourceDefinition, opts)
		}
	case ResourceQuark:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			quark, err := quarkClient.QuarkV1alpha1().Quarks(namespace).Get(
				context.TODO(), name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			applyAnnotationsAndLabels(quark, releaseName, namespace)
			return quarkClient.QuarkV1alpha1().Quarks(namespace).Update(
				context.TODO(), quark, metav1.UpdateOptions{})
		}
	default:
		return fmt.Errorf("unsupported resource type: %s", resource)
	}

	retryCount := 0
	for retryCount < 3 {
		_, err := updateFunc(ctx, metav1.UpdateOptions{})
		if err != nil {
			if apierrors.IsConflict(err) {
				log.Debugf("Resource conflict error, retrying (%d/3)...\n", retryCount+1)
				time.Sleep(2 * time.Second)
				retryCount++
			} else {
				return err
			}
		} else {
			break
		}
	}
	return nil
}

// applyAnnotationsAndLabels is a helper function that generically applies helm labels and annotations to a k8s resource
func applyAnnotationsAndLabels(obj metav1.Object, releaseName, namespace string) {
	labels := obj.GetLabels()
	if labels == nil {
		labels = make(map[string]string)
	}
	labels["app.kubernetes.io/managed-by"] = "Helm"
	obj.SetLabels(labels)

	annotations := obj.GetAnnotations()
	if annotations == nil {
		annotations = make(map[string]string)
	}
	annotations["meta.helm.sh/release-name"] = releaseName
	annotations["meta.helm.sh/release-namespace"] = namespace
	obj.SetAnnotations(annotations)
}

// CustomRESTClientGetter implements a clientcmd k8s network interface to convert a rest.Config to a connection endpoint for helm
type CustomRESTClientGetter struct {
	config     *rest.Config
	CacheDir   *string
	KubeConfig *string

	// config flags
	ClusterName        *string
	AuthInfoName       *string
	Context            *string
	Namespace          *string
	APIServer          *string
	TLSServerName      *string
	Insecure           *bool
	CertFile           *string
	CertData           *[]byte
	KeyFile            *string
	KeyData            *[]byte
	CAFile             *string
	CAData             *[]byte
	BearerToken        *string
	Impersonate        *string
	ImpersonateUID     *string
	ImpersonateGroup   *[]string
	Username           *string
	Password           *string
	Timeout            *string
	DisableCompression *bool
	// If non-nil, wrap config f
}

func NewCustomRESTClientGetter(c *rest.Config, namespace string) *CustomRESTClientGetter {

	tru := true

	customRest := &CustomRESTClientGetter{}
	customRest.config = c
	if c.CAFile != "" {
		customRest.CAFile = &c.CAFile
	}
	if c.CAData != nil && len(c.CAData) > 0 {
		customRest.CAData = &c.CAData
	}
	if c.CertFile != "" {
		customRest.CertFile = &c.CertFile
	}
	if c.CertData != nil && len(c.CertData) > 0 {
		customRest.CertData = &c.CertData
	}
	if c.KeyFile != "" {
		customRest.KeyFile = &c.KeyFile
	}
	if c.KeyData != nil && len(c.KeyData) > 0 {
		customRest.KeyData = &c.KeyData
	}
	if c.Host != "" {
		customRest.APIServer = &c.Host
	}
	if c.Insecure {
		customRest.Insecure = &tru
	}
	if c.TLSClientConfig.ServerName != "" {
		customRest.TLSServerName = &c.ServerName
	}
	if c.BearerToken != "" {
		customRest.BearerToken = &c.BearerToken
	}
	if c.Impersonate.UserName != "" {
		customRest.Impersonate = &c.Impersonate.UserName
	}
	if c.Impersonate.UID != "" {
		customRest.ImpersonateUID = &c.Impersonate.UID
	}
	if c.Impersonate.Groups != nil && len(c.Impersonate.Groups) > 0 {
		customRest.ImpersonateGroup = &c.Impersonate.Groups
	}
	if c.Username != "" {
		customRest.Username = &c.Username
	}
	if c.Password != "" {
		customRest.Password = &c.Password
	}
	if c.Timeout.Seconds() > 0 {
		timeoutStr := c.Timeout.String()
		customRest.Timeout = &timeoutStr
	}
	return customRest
}

func (c *CustomRESTClientGetter) ToRESTConfig() (*rest.Config, error) {
	return c.config, nil
}

func (f *CustomRESTClientGetter) ToRawKubeConfigLoader() clientcmd.ClientConfig {
	loadingRules := clientcmd.NewDefaultClientConfigLoadingRules()
	// use the standard defaults for this client command
	// DEPRECATED: remove and replace with something more accurate
	loadingRules.DefaultClientConfig = &clientcmd.DefaultClientConfig

	if f.KubeConfig != nil {
		loadingRules.ExplicitPath = *f.KubeConfig
	}

	overrides := &clientcmd.ConfigOverrides{ClusterDefaults: clientcmd.ClusterDefaults}
	if f.APIServer != nil {
		overrides.ClusterInfo.Server = *f.APIServer
	}
	// bind auth info flag values to overrides
	if f.CertFile != nil {
		overrides.AuthInfo.ClientCertificate = *f.CertFile
	}
	if f.CertData != nil {
		overrides.AuthInfo.ClientCertificateData = *f.CertData
	}
	if f.KeyFile != nil {
		overrides.AuthInfo.ClientKey = *f.KeyFile
	}
	if f.KeyData != nil {
		overrides.AuthInfo.ClientKeyData = *f.KeyData
	}
	if f.BearerToken != nil {
		overrides.AuthInfo.Token = *f.BearerToken
	}
	if f.Impersonate != nil {
		overrides.AuthInfo.Impersonate = *f.Impersonate
	}
	if f.ImpersonateUID != nil {
		overrides.AuthInfo.ImpersonateUID = *f.ImpersonateUID
	}
	if f.ImpersonateGroup != nil {
		overrides.AuthInfo.ImpersonateGroups = *f.ImpersonateGroup
	}
	if f.Username != nil {
		overrides.AuthInfo.Username = *f.Username
	}
	if f.Password != nil {
		overrides.AuthInfo.Password = *f.Password
	}

	// bind cluster flags
	if f.APIServer != nil {
		overrides.ClusterInfo.Server = *f.APIServer
	}
	if f.TLSServerName != nil {
		overrides.ClusterInfo.TLSServerName = *f.TLSServerName
	}
	if f.CAFile != nil {
		overrides.ClusterInfo.CertificateAuthority = *f.CAFile
	}
	if f.CAData != nil {
		overrides.ClusterInfo.CertificateAuthorityData = *f.CAData
	}
	if f.Insecure != nil {
		overrides.ClusterInfo.InsecureSkipTLSVerify = *f.Insecure
	}
	if f.DisableCompression != nil {
		overrides.ClusterInfo.DisableCompression = *f.DisableCompression
	}

	// bind context flags
	if f.Context != nil {
		overrides.CurrentContext = *f.Context
	}
	if f.ClusterName != nil {
		overrides.Context.Cluster = *f.ClusterName
	}
	if f.AuthInfoName != nil {
		overrides.Context.AuthInfo = *f.AuthInfoName
	}
	if f.Namespace != nil {
		overrides.Context.Namespace = *f.Namespace
	}

	if f.Timeout != nil {
		overrides.Timeout = *f.Timeout
	}

	// we only have an interactive prompt when a password is allowed
	if f.Password == nil {
		return clientcmd.NewNonInteractiveDeferredLoadingClientConfig(loadingRules, overrides)
	}
	return clientcmd.NewInteractiveDeferredLoadingClientConfig(loadingRules, overrides, os.Stdin)
}

func (c *CustomRESTClientGetter) ToDiscoveryClient() (discovery.CachedDiscoveryInterface, error) {
	discoveryClient, err := discovery.NewDiscoveryClientForConfig(c.config)
	if err != nil {
		return nil, err
	}
	return memory.NewMemCacheClient(discoveryClient), nil
}

func (c *CustomRESTClientGetter) ToRESTMapper() (meta.RESTMapper, error) {
	discoveryClient, err := c.ToDiscoveryClient()
	if err != nil {
		return nil, err
	}

	mapper := restmapper.NewDeferredDiscoveryRESTMapper(discoveryClient)
	expander := restmapper.NewShortcutExpander(mapper, discoveryClient)
	return expander, nil
}
